use std::{cell::RefCell, collections::HashMap, rc::Rc};

use crate::{
    DeBrujinAst,
    inference::{
        recursive::infer_type as infer_type_rec, unification::Constraints,
        unification::infer_type as infer_type_uni,
    },
    multi_map::MultiMap,
    parse::{Ast, Constant},
    types::{PrimitiveType, TaggedType, Type, TypeTag},
    vec_map::VecMap,
};

use super::unification::{TypeVarAst, TypeVarCtx};

#[test]
fn infer_add_nat_rec() {
    let ast: DeBrujinAst = Ast::Application(
        Box::new(Ast::Application(
            Box::new(Ast::Variable("add".to_string())),
            Box::new(Ast::Constant(Constant::Nat(5))),
        )),
        Box::new(Ast::Constant(Constant::Nat(7))),
    )
    .into();

    let mut gamma = HashMap::new();
    gamma.insert(
        "add".to_string(),
        crate::types::TaggedType::Tagged(
            TypeTag::Num,
            "a".to_string(),
            Box::new(TaggedType::Concrete(Type::Arrow(
                Box::new(Type::TypeVariable("a".to_string())),
                Box::new(Type::Arrow(
                    Box::new(Type::TypeVariable("a".to_string())),
                    Box::new(Type::TypeVariable("a".to_string())),
                )),
            ))),
        ),
    );

    let infered = infer_type_rec(&gamma, ast).unwrap();
    assert_eq!(
        infered,
        TaggedType::Concrete(Type::Primitive(PrimitiveType::Nat))
    );
}

#[test]
fn infer_add_nat_partial_rec() {
    let ast: DeBrujinAst = Ast::Application(
        Box::new(Ast::Variable("add".to_string())),
        Box::new(Ast::Constant(Constant::Nat(5))),
    )
    .into();

    let mut gamma = HashMap::new();
    gamma.insert(
        "add".to_string(),
        crate::types::TaggedType::Tagged(
            TypeTag::Num,
            "a".to_string(),
            Box::new(TaggedType::Concrete(Type::Arrow(
                Box::new(Type::TypeVariable("a".to_string())),
                Box::new(Type::Arrow(
                    Box::new(Type::TypeVariable("a".to_string())),
                    Box::new(Type::TypeVariable("a".to_string())),
                )),
            ))),
        ),
    );

    let infered = infer_type_rec(&gamma, ast).unwrap();
    assert_eq!(
        infered,
        TaggedType::Concrete(Type::arrow(PrimitiveType::Nat, PrimitiveType::Nat))
    );
}

#[test]
fn subtype_constraints() {
    let typ = TaggedType::Tagged(
        TypeTag::Num,
        "a".to_string(),
        Box::new(Type::arrow("a", "a").into()),
    );

    let mut st_constraints = Rc::new(RefCell::new(MultiMap::new()));
    let ctx = TypeVarCtx::new();

    let res = typ.make_constraints(st_constraints.clone(), &ctx);

    assert_eq!(
        st_constraints
            .borrow()
            .get(&Type::TypeVariable("?typ_0".to_string())),
        vec![&TypeTag::Num]
    );

    assert_eq!(res, Type::arrow("?typ_0", "?typ_0"));
}

#[test]
fn infer_add_nat_uni() {
    let ast: DeBrujinAst = Ast::Application(
        Box::new(Ast::Application(
            Box::new(Ast::Variable("add".to_string())),
            Box::new(Ast::Constant(Constant::Nat(5))),
        )),
        Box::new(Ast::Constant(Constant::Nat(7))),
    )
    .into();

    let mut gamma = HashMap::new();
    gamma.insert(
        "add".to_string(),
        TaggedType::Tagged(
            TypeTag::Num,
            "a".to_string(),
            Box::new(TaggedType::Concrete(Type::arrow(
                "a",
                Type::arrow("a", "a"),
            ))),
        ),
    );

    let infered = infer_type_uni(&gamma, ast).unwrap();
    assert_eq!(
        infered,
        TaggedType::Concrete(Type::Primitive(PrimitiveType::Nat))
    );
}

#[test]
fn infer_add_nat_partial_uni() {
    let ast: DeBrujinAst = Ast::Application(
        Box::new(Ast::Variable("add".to_string())),
        Box::new(Ast::Constant(Constant::Nat(5))),
    )
    .into();

    let mut gamma = HashMap::new();
    gamma.insert(
        "add".to_string(),
        TaggedType::Concrete(Type::arrow(
            PrimitiveType::Nat,
            Type::arrow(PrimitiveType::Nat, PrimitiveType::Nat),
        )),
    );

    let typ = infer_type_uni(&gamma, ast).unwrap();

    assert_eq!(
        typ,
        TaggedType::Concrete(Type::arrow(PrimitiveType::Nat, PrimitiveType::Nat))
    );
}

#[test]
fn infer_id_uni() {
    let ast: DeBrujinAst = Ast::Abstraction(
        "x".to_string(),
        None,
        Box::new(Ast::Variable("x".to_string())),
    )
    .into();

    let mut gamma = HashMap::new();
    let typ = infer_type_uni(&gamma, ast).unwrap();

    assert_eq!(
        typ,
        TaggedType::Tagged(
            TypeTag::Any,
            "?typ_1".to_string(),
            Box::new(TaggedType::Concrete(Type::arrow("?typ_1", "?typ_1")))
        )
    );
}

#[test]
fn subst_type_var() {
    let typ = TaggedType::Tagged(
        TypeTag::Num,
        "a".to_string(),
        Box::new(TaggedType::Concrete(Type::arrow("a", "?typ_4"))),
    );

    let subst = TaggedType::Tagged(
        TypeTag::Any,
        "b".to_string(),
        Box::new(TaggedType::Concrete(Type::arrow("b", "b"))),
    );

    let typ = typ.subst_typevar("?typ_4", 0, subst).unwrap();

    assert_eq!(
        typ,
        TaggedType::Tagged(
            TypeTag::Num,
            "a".to_string(),
            Box::new(TaggedType::Tagged(
                TypeTag::Any,
                "b".to_string(),
                Box::new(TaggedType::Concrete(Type::arrow(
                    "a",
                    Type::arrow("b", "b")
                )))
            ))
        )
    );

    let typ = TaggedType::Tagged(
        TypeTag::Num,
        "a".to_string(),
        Box::new(TaggedType::Concrete(Type::arrow("a", "?typ_4"))),
    );

    let subst = TaggedType::Tagged(
        TypeTag::Any,
        "a".to_string(),
        Box::new(TaggedType::Concrete(Type::arrow("a", "a"))),
    );

    let typ = typ.subst_typevar("?typ_4", 0, subst).unwrap();

    assert!(match typ {
        TaggedType::Tagged(TypeTag::Num, a, tagged_type) => match *tagged_type {
            TaggedType::Tagged(TypeTag::Any, b, tagged_type) => match *tagged_type {
                TaggedType::Concrete(typ) => match typ {
                    Type::Arrow(lhs, rhs) => match (*lhs, *rhs) {
                        (Type::TypeVariable(a1), Type::Arrow(b1, b2)) if a1 == a => {
                            match (*b1, *b2) {
                                (Type::TypeVariable(b1), Type::TypeVariable(b2)) => {
                                    (b1 == b) && (b2 == b)
                                }
                                _ => false,
                            }
                        }
                        _ => false,
                    },
                    _ => false,
                },
                _ => false,
            },
            _ => false,
        },
        _ => false,
    });
}
